@@ -12,10 +12,200 @@ from . import __project_name__, _errors
 from . import _debug  # noqa:F401 isort:skip
 
 
+DEFAULT_CONFIG_FILEPATH = os.path.join(xdg_config_home, __project_name__, 'config.cfg')
 DEFAULT_LOCATIONS_FILEPATH = os.path.join(xdg_config_home, __project_name__, 'locations')
 DEFAULT_CLIENTS_FILEPATH = os.path.join(xdg_config_home, __project_name__, 'clients.ini')
 
 
+def _as_dict(cfg):
+    # Convert ConfigParser object to dict
+    return {
+        section: dict(cfg.items(section))
+        for section in cfg.sections()
+    }
+
+
+def _convert_to_client(options, option):
+    client_name = options[option]
+    if client_name not in aiobtclientapi.client_names():
+        raise ValueError(f'{client_name}: Unknown client')
+    return client_name
+
+
+def _convert_to_url(options, option):
+    url = options[option]
+    client_name = options['client']
+    client_cls = getattr(aiobtclientapi, f'{client_name.capitalize()}API')
+    try:
+        return client_cls.URL(url)
+    except ValueError as e:
+        raise ValueError(f'{url}: {e}')
+
+
+_bool_values = {
+    '1': True, 'yes': True, 'true': True, 'on': True,
+    '0': False, 'no': False, 'false': False, 'off': False,
+}
+
+def _convert_to_bool(options, option):
+    value = str(options[option]).lower()
+    try:
+        return _bool_values[value]
+    except KeyError:
+        raise ValueError(f'{value}: Must be true/false, yes/no, on/off, 1/0')
+
+
+def _convert_to_string(options, option):
+    return str(options[option]).replace('\n', ' ')
+
+
+def _convert_to_list(options, option):
+    value = str(options[option]).strip()
+    if value:
+        return value.split('\n')
+    else:
+        return []
+
+
+def _convert_to_umask(options, option):
+    value = options[option]
+    if value is None:
+        return None
+    elif not value:
+        return 0
+    elif re.search(r'^[0-7]{1,4}$', str(value)):
+        return int(value, 8)
+    else:
+        raise ValueError(f'{value}: Invalid umask')
+
+
+class Config(collections.abc.Mapping):
+    """
+    :class:`dict` subclass that reads simple `option = value` lines from file
+
+    Values may contain line breaks if there is some whitespace after each line
+    break. Use :func:`_convert_to_list` or :func:`_convert_to_string` to handle
+    line breaks.
+
+    :param filepath: File to read configuration from
+
+    :raise ConfigError: if reading or parsing `filepath` fails
+    """
+
+    def __init__(self, filepath):
+        self._filepath = filepath
+        self._config = self._read(filepath)
+
+    def _read(self, filepath):
+        try:
+            text = open(filepath, 'r').read()
+        except OSError as e:
+            # Ignore missing default config file path
+            if e.errno != errno.ENOENT or filepath != DEFAULT_CONFIG_FILEPATH:
+                msg = e.strerror if e.strerror else str(e)
+                raise _errors.ConfigError(f'Failed to read: {msg}', filepath=filepath)
+            else:
+                text = ''
+
+        # Validation
+        dct = self._parse(text, filepath)
+        self._ensure_only_valid_options(dct, filepath)
+        self._fill_in_defaults(dct)
+        self._convert_values(dct, filepath)
+        return dct
+
+    def _parse(self, text, filepath):
+        dct = {}
+        current_option = None
+
+        for i, line in enumerate(re.split('[\n\r]', text), start=1):
+            # comment or empty line
+            if re.search(r'^\s*#', line) or not line.strip():
+                continue
+
+            # option = value
+            match = re.search(r'^\s*(\w+)\s*=\s*(.*?)\s*$', line)
+            if match:
+                dct[match.group(1)] = match.group(2)
+                current_option = match.group(1)
+                continue
+
+            # option = value1\n value2
+            # or
+            # option = \n value1\n value2
+            match = re.search(r'^\s+(.*?)\s*$', line)
+            if match and current_option is not None:
+                if match.group(1).strip():
+                    if dct[current_option]:
+                        dct[current_option] += '\n' + match.group(1)
+                    else:
+                        dct[current_option] = match.group(1)
+                continue
+
+            raise _errors.ConfigError(f'{line}: Invalid syntax',
+                                      filepath=filepath, line_number=i)
+
+        return dct
+
+    _valid_options = (
+        'before_location_search_commands',
+        'after_location_found_commands',
+        'after_torrent_handled_commands',
+        'umask',
+        'copy_torrent_to',
+    )
+
+    def _ensure_only_valid_options(self, dct, filepath):
+        for option in dct:
+            if option not in self._valid_options:
+                raise _errors.ConfigError(f'Unknown option: {option}', filepath=filepath)
+
+    _defaults = {
+        'before_location_search_commands': '',
+        'after_location_found_commands': '',
+        'after_torrent_handled_commands': '',
+        'umask': None,
+        'copy_torrent_to': '',
+    }
+
+    def _fill_in_defaults(self, dct):
+        for option in self._defaults:
+            if option not in dct:
+                dct[option] = self._defaults[option]
+
+    _convert_values_map = {
+        'before_location_search_commands': _convert_to_list,
+        'after_location_found_commands': _convert_to_list,
+        'after_torrent_handled_commands': _convert_to_list,
+        'umask': _convert_to_umask,
+        'copy_torrent_to': _convert_to_string,
+    }
+
+    def _convert_values(self, dct, filepath):
+        for option, converter in self._convert_values_map.items():
+            try:
+                dct[option] = converter(dct, option)
+            except ValueError as e:
+                raise _errors.ConfigError(f'{option}: {e}', filepath=filepath)
+
+    @property
+    def filepath(self):
+        """Where the configuration was read from"""
+        return self._filepath
+
+    def __getitem__(self, key):
+        return self._config[key]
+
+    def __len__(self):
+        return len(self._config)
+
+    def __iter__(self):
+        return iter(self._config)
+
+    def __repr__(self):
+        return f'<{type(self).__name__} {self._filepath!r} {self._config!r}>'
+
+
 class Locations(collections.abc.MutableSequence):
     """
     :class:`list` subclass that reads directory paths from `filepath`
@@ -181,8 +371,7 @@ class Clients(collections.abc.Mapping):
 
     :param filepath: File to read client configurations from
 
-    :raise ConfigError: if reading `filepath` fails or if it contains a file
-        path
+    :raise ConfigError: if reading or parsing `filepath` fails
     """
 
     def __init__(self, filepath):
@@ -220,19 +409,13 @@ class Clients(collections.abc.Mapping):
             raise _errors.ConfigError(str(e), filepath=filepath)
 
         # Validation
-        dct = self._as_dict(cfg)
-        self._ensure_only_valid_options(filepath, dct)
-        self._ensure_mandatory_options(filepath, dct)
+        dct = _as_dict(cfg)
+        self._ensure_only_valid_options(dct, filepath)
+        self._ensure_mandatory_options(dct, filepath)
         self._fill_in_defaults(dct)
-        self._convert_values(filepath, dct)
+        self._convert_values(dct, filepath)
         return dct
 
-    def _as_dict(self, cfg):
-        return {
-            section: dict(cfg.items(section))
-            for section in cfg.sections()
-        }
-
     _valid_options = (
         'client',
         'url',
@@ -242,7 +425,7 @@ class Clients(collections.abc.Mapping):
         'stopped',
     )
 
-    def _ensure_only_valid_options(self, filepath, dct):
+    def _ensure_only_valid_options(self, dct, filepath):
         for section in dct:
             for option in dct[section]:
                 if option not in self._valid_options:
@@ -250,7 +433,7 @@ class Clients(collections.abc.Mapping):
 
     _mandatory_options = ('client',)
 
-    def _ensure_mandatory_options(self, filepath, dct):
+    def _ensure_mandatory_options(self, dct, filepath):
         for section in dct:
             for option in self._mandatory_options:
                 if option not in dct[section]:
@@ -291,12 +474,14 @@ class Clients(collections.abc.Mapping):
                     else:
                         dct[section][option] = self._defaults[option]
 
-    def _convert_values(self, filepath, dct):
+    def _convert_values(self, dct, filepath):
         convert_map = {
-            'client': self._convert_to_client,
-            'url': self._convert_to_url,
-            'verify': self._convert_to_bool,
-            'stopped': self._convert_to_bool,
+            'client': _convert_to_client,
+            'url': _convert_to_url,
+            'username': _convert_to_string,
+            'password': _convert_to_string,
+            'verify': _convert_to_bool,
+            'stopped': _convert_to_bool,
         }
         for section, options in dct.items():
             for option, converter in convert_map.items():
@@ -305,33 +490,6 @@ class Clients(collections.abc.Mapping):
                 except ValueError as e:
                     raise _errors.ConfigError(f'{section}: {option}: {e}', filepath=filepath)
 
-    def _convert_to_client(self, options, option):
-        client_name = options[option]
-        if client_name not in aiobtclientapi.client_names():
-            raise ValueError(f'{client_name}: Unknown client')
-        return client_name
-
-    def _convert_to_url(self, options, option):
-        url = options[option]
-        client_name = options['client']
-        client_cls = getattr(aiobtclientapi, f'{client_name.capitalize()}API')
-        try:
-            return client_cls.URL(url)
-        except ValueError as e:
-            raise ValueError(f'{url}: {e}')
-
-    _bool_values = {
-        '1': True, 'yes': True, 'true': True, 'on': True,
-        '0': False, 'no': False, 'false': False, 'off': False,
-    }
-
-    def _convert_to_bool(self, options, option):
-        value = str(options[option]).lower()
-        try:
-            return self._bool_values[value]
-        except KeyError:
-            raise ValueError(f'{value}: Must be true/false, yes/no, on/off, 1/0')
-
     @property
     def filepath(self):
         """Where the configuration was read from"""
